import bcrypt


def encrypt_password(password):
    if type(password) is str:
        password = password.encode()
    hashed = bcrypt.hashpw(password, bcrypt.gensalt())
    return hashed.decode()


def password_matches(user, password):
    hashed = user.password
    if type(hashed) is str:
        hashed = hashed.encode()
    if type(password) is str:
        password = password.encode()
    return bcrypt.checkpw(password, hashed)



import os
import types

import transaction
from pyramid.httpexceptions import HTTPNotFound
from querystring_parser import parser as qsparser
from hornstone.alchemy import TimeStampMixin


def apiroot(prefix='/api', version='dev'):
    return os.path.join(prefix, version)


class BaseResource(object):
    def __init__(self, request, context=None):
        self.request = request
        self.db = self.request.dbsession
        self.limit = 25
        self.max_limit = 100
        self._use_pagination = True

    def serialize_object(self, dbobj):
        return dbobj.serialize()

    # use this when you do only need  certain model attributes
    # in a collection, rather than return the complete models
    def serialize_object_for_collection_query(self, dbobj):
        # FIXME test this better
        if hasattr(dbobj, 'keys') and isinstance(dbobj.keys, types.MethodType):
            return dict(((k, getattr(dbobj, k)) for k in dbobj.keys()))
        else:
            return self.serialize_object(dbobj)

    def collection_query(self):
        raise RuntimeError("Implement me in subclass")

    def _apply_pagination(self, query):
        offset = 0
        limit = self.limit
        GET = self.request.GET
        if 'offset' in GET:
            offset = int(GET['offset'])
        if 'limit' in GET:
            limit = int(GET['limit'])
            if limit > self.max_limit:
                limit = self.max_limit
        return query.offset(offset).limit(limit)

    def collection_get(self):
        qs = qsparser.parse(self.request.query_string)
        if 'columns' in qs and '' in qs['columns'] and len(qs['columns']['']):
            _fields = (getattr(self.model, f) for f in qs['columns'][''])
            query = self.db.query(*_fields)
        else:
            query = self.collection_query()

        if 'distinct' in qs:
            # setup distinct and groupby
            raise RuntimeError("No support for distinct yet.")
        if 'where' in qs:
            # setup where clauses
            raise RuntimeError("No support for where yet.")

        q = query
        total_count = q.count()

        if self._use_pagination:
            q = self._apply_pagination(q)
        data = [self.serialize_object_for_collection_query(o) for o in q]
        return dict(total_count=total_count, items=data)


# from bookshelf-csapi
#
# req.query.withRelated - array of related models
# req.query.columns - array of columns
# req.query.where - where clause (two formats?)
# req.query.distinct - distinct values
# req.query.offset
# req.query.limit
# req.query.sort - only one column? always order by key unless present?
# req.query.direction - default is 'asc'
#
# if sort or offset is present set the direction
#
#
#
#

# BaseResource.model must be set in subclass
class BaseModelResource(BaseResource):
    def _isTimeStampMixin(self):
        return isinstance(self.model(), TimeStampMixin)

    def collection_query(self):
        return self.db.query(self.model)

    def collection_post(self):
        with transaction.manager:
            m = self.model()
            for field in self.request.json:
                value = self.request.json[field]
                if type(value) is dict:
                    print("value of field {} is dict".format(field))
                setattr(m, field, value)
            # FIXME
            if hasattr(m, 'user_id'):
                m.user_id = self.request.user.id
            self.db.add(m)
            self.db.flush()
        return self.serialize_object(m)

    def get(self):
        id = self.request.matchdict['id']
        m = self.db.query(self.model).get(id)
        if m is None:
            raise HTTPNotFound
        return self.serialize_object(m)

    def put(self):
        with transaction.manager:
            id = self.request.matchdict['id']
            m = self.db.query(self.model).get(id)
            if m is None:
                raise HTTPNotFound
            fields = set(self.request.json.keys())
            if self._isTimeStampMixin():
                fields.discard('created')
                fields.discard('updated')
            for field in fields:
                setattr(m, field, self.request.json[field])
            self.db.add(m)
            self.db.flush()
        return self.serialize_object(m)

    def delete(self):
        with transaction.manager:
            id = self.request.matchdict['id']
            m = self.db.query(self.model).get(id)
            if m is None:
                raise HTTPNotFound
            self.db.delete(m)

    @property
    def model(self):
        raise NotImplementedError


class SimpleModelResource(BaseModelResource):
    """Handles /api/{model}[/{id}]"""
    @property
    def model(self):
        return self.model_map.get(self.request.matchdict['model'])

    @property
    def model_map(self):
        raise NotImplementedError





from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import FileResponse
from pyramid.path import AssetResolver


def static_asset_response(request, asset):
    resolver = AssetResolver()
    descriptor = resolver.resolve(asset)
    if not descriptor.exists():
        raise HTTPNotFound(request.url)
    path = descriptor.abspath()
    response = FileResponse(path, request)
    zip_response = False
    for ending in ['.css', '.js', '.coffee', '.html', '.ttf']:
        if path.endswith(ending):
            zip_response = True
    if zip_response:
        response.encode_content()
    for ending in ['.css', '.js']:
        # one day for css and js
        if path.endswith(ending):
            response.cache_expires(3600 * 24)
            response.cache_control.public = True
    if path.endswith('.ttf'):
        # one year for fonts
        response.cache_expires(3600 * 24 * 365)
        response.cache_control.public = True
    return response


class BaseView(object):
    def __init__(self, request):
        self.request = request

    def get_app_settings(self):
        return self.request.registry.settings


class BaseUserView(BaseView):
    def get_current_user(self):
        "Get user db object"
        return self.request.user


class BaseViewCallable(BaseView):
    def __init__(self, request):
        super(BaseViewCallable, self).__init__(request)
        self.response = None
        self.data = {}

    def __call__(self):
        if self.response is not None:
            return self.response
        else:
            return self.data


class BaseUserViewCallable(BaseViewCallable, BaseUserView):
    pass



from pyramid.view import view_defaults

from .base import BaseViewCallable


@view_defaults(renderer='trumpet:templates/mainview.mako')
class BaseClientView(BaseViewCallable):
    def __init__(self, request):
        super(BaseClientView, self).__init__(request)
        self.settings = self.get_app_settings()
        self.data = dict()



from sqlalchemy.orm.exc import NoResultFound
import transaction


class BaseManager(object):
    def __init__(self, session):
        self.session = session

    def set_session(self, session):
        self.session = session

    def _range_filter(self, query, column, start, end):
        query = query.filter(column >= start)
        query = query.filter(column <= end)
        return query

    @property
    def query(self):
        # return self.session.query(self.dbmodel)
        raise NotImplementedError("Override me")

    def get(self, id):
        return self.query.get(id)

    def all(self):
        return self.query.all()

    def delete_everything(self):
        self.query.delete()

    def delete_everything_tm(self):
        with transaction.manager:
            self.delete_everything()


class GetByNameManager(BaseManager):
    def get_by_name_query(self, name):
        return self.query.filter_by(name=name)

    def get_by_name(self, name):
        q = self.get_by_name_query(name)
        try:
            return q.one()
        except NoResultFound:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'hornstone',
    'pyramid >= 1.9a',
    'pyramid_tm',
    'plaster_pastedeploy',
    'zope.sqlalchemy',
    'pyramid-beaker',  # session management
    'pyramid-mako',
    'cornice',         # REST views
    'waitress',
    'bcrypt',
    # testing below
    'querystring_parser',
    # 'versiontools',
]

tests_require = [
    'WebTest >= 1.3.1',  # py3 compat
    'pytest',
    'pytest-cov',
]

setup_requirements = [
    'pytest-runner',
    # TODO(umeboshi2): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'WebTest >= 1.3.1',  # py3 compat
    'pytest',
    'pytest-cov',
    # TODO: put package test requirements here
]

setup(
    name='trumpet',
    version='0.2.3',
    description="Build a website with pyramid",
    long_description=readme + '\n\n' + history,
    author="Joseph Rawson",
    author_email='joseph.rawson.works@gmail.com',
    url='https://github.com/umeboshi2/trumpet',
    packages=find_packages(include=['trumpet', 'trumpet.*']),
    entry_points={
        'paste.app_factory': [
            'main = trumpet:main',
        ],
        'console_scripts': [
            'initialize_trumpet_db = trumpet.scripts.initializedb:main',
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="UNLICENSED",
    zip_safe=False,
    keywords='trumpet',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: Public Domain',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



from datetime import datetime


def _parse_date(date):
    year, month, day = [int(p) for p in date.split('-')]
    return datetime(year, month, day)


def get_start_end_from_request(request, timestamps=False,
                               verbose=False):
    start = request.GET['start']
    end = request.GET['end']
    if verbose:
        print("START, END", start, end)
    if not timestamps:
        start = _parse_date(start)
        end = _parse_date(end)
    return start, end



from crypt import crypt
import random
from string import ascii_letters

from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authentication import SessionAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from sqlalchemy.orm.exc import NoResultFound


def make_salt(id=5, length=10):
    phrase = ''
    for ignore in range(length):
        phrase += random.choice(ascii_letters)
    return '$%d$%s$' % (id, phrase)


def encrypt_password(password):
    salt = make_salt()
    encrypted = crypt(password, salt)
    return encrypted


def check_password(encrypted, password):
    if '$' not in encrypted:
        raise RuntimeError("we are supposed to be using random salt.")
    ignore, id, salt, epass = encrypted.split('$')
    salt = '$%s$%s$' % (id, salt)
    check = crypt(password, salt)
    return check == encrypted


def authenticate(userid, request):
    db = request.db
    usermodel = request.usermodel
    fieldname = request.registry.settings['db.usernamefield']
    field = getattr(usermodel, fieldname)
    try:
        user = db.query(usermodel).filter(field == userid).one()
    except NoResultFound:
        return
    if user is not None:
        return user.get_groups()


def check_user_password(user, password):
    return check_password(user.pw.password, password)


authn_policy = SessionAuthenticationPolicy(callback=authenticate)
authz_policy = ACLAuthorizationPolicy()


def make_authn_policy(secret, cookie, callback=authenticate,
                      timeout=None):
    ap = AuthTktAuthenticationPolicy(
        secret=secret,
        callback=callback,
        cookie_name=cookie,
        timeout=timeout)
    return ap


def make_session_authn_policy(callback=authenticate):
    sp = SessionAuthenticationPolicy(callback=callback,
                                     )
    return sp


def make_authn_authz_policies(secret, cookie, callback=authenticate,
                              timeout=None, tkt=True):
    if tkt:
        authn_policy = make_authn_policy(secret, cookie, callback, timeout)
    else:
        authn_policy = make_session_authn_policy(callback=callback)
    authz_policy = ACLAuthorizationPolicy()
    return authn_policy, authz_policy



from pyramid.security import Allow, Everyone, Authenticated


class Resource(dict):
    def __init__(self, name, parent):
        self.__name__ = name
        self.__parent__ = parent


class ResourceObject(object):
    def __init__(self, name, parent=None):
        self.__name__ = name
        self.__parent__ = parent

# the acl entries are allow/deny, group, permission


class RootGroupFactory(Resource):
    __default_acl__ = [
        (Allow, Everyone, 'public'),
        (Allow, Authenticated, 'user'),
        (Allow, 'manager', 'manage'),
        (Allow, 'admin', ('admin', 'manage')),
    ]
    authn_policy = None

    @property
    def __acl__(self):
        return self.__default_acl__

    def __init__(self, request):
        self.request = request
        # self['admin'] = dict()


# straight from deform docs
class MemoryTmpStore(dict):
    def preview_url(self, name):
        return None
